"""Views for the marion application"""

import json

from django.conf import settings
from django.core.exceptions import PermissionDenied, ValidationError
from django.http import HttpResponse, HttpResponseBadRequest
from django.template import Context
from django.utils.module_loading import import_string

from rest_framework import mixins, status, viewsets
from rest_framework.response import Response

from .exceptions import (
    DocumentIssuerContextQueryValidationError,
    DocumentIssuerContextValidationError,
)
from .models import DocumentRequest
from .serializers import DocumentRequestSerializer


class DocumentRequestViewSet(
    mixins.CreateModelMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
):
    """API endpoint that allows document requests to be viewed, listed or created"""

    # pylint: disable=too-many-ancestors
    queryset = DocumentRequest.objects.all()
    serializer_class = DocumentRequestSerializer

    def create(self, request, *args, **kwargs):
        """Create a document request (and the corresponding document)"""

        try:
            return super().create(request, *args, **kwargs)
        except (
            DocumentIssuerContextQueryValidationError,
            DocumentIssuerContextValidationError,
            ValidationError,
        ) as error:
            return Response(
                data={"error": str(error)}, status=status.HTTP_400_BAD_REQUEST
            )


def document_template_debug(request):
    """Document template debug view.

    Disclaimer: this view should be used for development/testing purpose only.
    """

    if not settings.DEBUG:
        raise PermissionDenied

    issuer_path = request.GET.get("issuer", None)
    context = json.loads(request.GET.get("context", "{}"))

    if issuer_path is None:
        return HttpResponseBadRequest("You should provide an issuer.")

    try:
        issuer = import_string(issuer_path)()
    except ImportError:
        return HttpResponseBadRequest(f"Unknown issuer {issuer_path}.")

    css = issuer.get_css().render(Context(context))
    context.update({"css": css, "debug": True})
    html = issuer.get_html().render(Context(context))

    return HttpResponse(html)



"""Tests for the marion application defaults"""

import importlib
from pathlib import Path

from marion import defaults


def test_defaults_overrides_with_settings(settings):
    """Test marion.defaults overrides with settings definition"""

    settings.MARION_DOCUMENT_ISSUER_CHOICES_CLASS = (
        "howard.documents.DocumentIssuerChoices"
    )
    settings.MARION_DOCUMENTS_ROOT = Path("/tmp/documents/abc")
    settings.MARION_DOCUMENTS_TEMPLATE_ROOT = Path("howard/documents/abc")

    # Force module reload to take into account setting override as it is loaded
    # very early in the stack
    importlib.reload(defaults)

    assert (
        defaults.DOCUMENT_ISSUER_CHOICES_CLASS
        == "howard.documents.DocumentIssuerChoices"
    )
    assert defaults.DOCUMENTS_ROOT == Path("/tmp/documents/abc")
    assert defaults.DOCUMENTS_TEMPLATE_ROOT == Path("howard/documents/abc")


def teardown_function():
    """Executed at the end of the current test suite"""

    # Force module reload as the default test settings have been restored
    importlib.reload(defaults)



"""Tests for the marion.utils module"""

from pathlib import Path

from django.conf import settings

import marion
from marion.utils import static_file_fetcher


# pylint: disable=invalid-name
def test_static_file_fetcher(fs):
    """Test weasyprint custom static file fetcher."""

    # Create a fake static file for the marion app
    relative_static_file_path = "marion/test.txt"
    static_file_path = Path(f"{marion.__path__[0]}/static/{relative_static_file_path}")
    fs.create_file(static_file_path, contents="This is content.")

    # Fetch this file
    data = static_file_fetcher(
        f"file://{settings.STATIC_URL}{relative_static_file_path}"
    )
    assert data.get("filename") == static_file_path.name
    assert data.get("mime_type") == "text/plain"
    file_ = data.get("file_obj")
    assert file_.readlines() == [
        b"This is content.",
    ]
    file_.close()

    # Test Weasyprint fallback (if the file is not located in a Django static file path)
    file_path = "/tmp/foo.txt"
    fs.create_file(file_path, contents="This is a different content.")

    data = static_file_fetcher(f"file://{file_path}")
    assert data.get("mime_type") == "text/plain"
    file_ = data.get("file_obj")
    assert file_.readlines() == [
        b"This is a different content.",
    ]
    file_.close()



"""Dummy Document"""

from uuid import UUID

from pydantic import BaseModel, constr

from .base import AbstractDocument


class ContextModel(BaseModel):
    """Context model definition"""

    identifier: UUID
    fullname: constr(min_length=2, max_length=255)

    class Config:
        """Context model configuration"""

        extra = "forbid"


class ContextQueryModel(BaseModel):
    """ContextQuery model definition"""

    fullname: constr(min_length=2, max_length=255)

    class Config:
        """ContextQuery model configuration"""

        extra = "forbid"


class DummyDocument(AbstractDocument):
    """A test document for marion"""

    context_model = ContextModel

    context_query_model = ContextQueryModel

    keywords = ["dummy", "test", "document"]

    def fetch_context(self) -> dict:
        """Fetch the context that will be used to compile the document template.

        This method is required by the AbstractDocument interface. The
        context query parameters should be sufficient to get template context.

        """

        return {"fullname": self.context_query.fullname, "identifier": self.identifier}

    def get_title(self):
        """Define a custom title for the generated PDF metadata"""

        return f"Dummy document {self.identifier}"



"""Factories for the marion application"""

from factory.django import DjangoModelFactory

from . import models


class DocumentRequestFactory(DjangoModelFactory):
    """DocumentRequest factory"""

    class Meta:  # noqa
        model = models.DocumentRequest



"""Documents generation utilities."""

import mimetypes
from pathlib import Path
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.staticfiles.finders import find
from django.core.exceptions import SuspiciousFileOperation

import weasyprint


def static_file_fetcher(url, *args, **kwargs):
    """Weasyprint static files fetcher.

    If the file URL starts with 'file://', it will be fetched from the configured
    storage.

    The following code has been adapted from the django-weasyprint project [1].

    References:

    [1] https://github.com/fdemmer/django-weasyprint/
    """

    if url.startswith("file:"):
        mime_type, encoding = mimetypes.guess_type(url)
        url_path = urlparse(url).path
        data = {
            "mime_type": mime_type,
            "encoding": encoding,
            "filename": Path(url_path).name,
        }

        path = url_path.replace(settings.STATIC_URL, "", 1)
        try:
            data["file_obj"] = open(find(path), "rb")
        # A SuspiciousFileOperation is raised by Django if the file has been
        # found outside referenced static file paths. In this case, we ignore
        # this error and fallback to the default Weasyprint fetcher for files
        # referenced with the file:// protocol.
        except SuspiciousFileOperation:
            pass
        else:
            return data

    # Fall back to weasyprint default fetcher
    return weasyprint.default_url_fetcher(url, *args, **kwargs)



"""Marion, the documents factory"""

import importlib.metadata
from pathlib import Path

from setuptools.config import read_configuration


def _get_version():
    """Get version from installed package with a fallback to the setup.cfg version
    string.

    """

    try:
        return importlib.metadata.version("marion")
    except importlib.metadata.PackageNotFoundError:
        return read_configuration(Path(__file__).parent / ".." / "setup.cfg")[
            "metadata"
        ]["version"]


__version__ = _get_version()



"""Default settings for the marion application.

Application settings should always be imported from here to ensure that:

1. they can be overriden in the Django project settings using the following
naming pattern: MARION_{{ << setting name >> }}, _e.g._ MARION_FOO for the FOO
setting of marion.

2. it has relevant, properly maintained defaults.

"""

from pathlib import Path

from django.conf import settings
from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _

DOCUMENT_ISSUER_CHOICES_CLASS = getattr(
    settings,
    "MARION_DOCUMENT_ISSUER_CHOICES_CLASS",
    "marion.defaults.DocumentIssuerChoices",
)
DOCUMENTS_ROOT = getattr(settings, "MARION_DOCUMENTS_ROOT", Path(settings.MEDIA_ROOT))
DOCUMENTS_TEMPLATE_ROOT = getattr(
    settings, "MARION_DOCUMENTS_TEMPLATE_ROOT", Path("marion")
)


class DocumentIssuerChoices(TextChoices):
    """Active document issuers.

    Note that changing the default choices for the list of allowed document
    issuers may require to create and apply a database migration.

    """

    DUMMY = "marion.issuers.DummyDocument", _("Dummy")



"""Debug Urls for the marion application"""

from django.urls import path

from .. import views

urlpatterns = [
    path(
        "templates/",
        views.document_template_debug,
        name="documents-template-debug",
    )
]



"""Base document issuer for the marion application"""

import uuid
from abc import ABC, abstractmethod
from typing import Union

from django.conf import settings
from django.template import Context
from django.template.engine import Engine
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.text import re_camel_case
from django.utils.translation import gettext_lazy as _

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError
from weasyprint import CSS, HTML
from weasyprint.document import DocumentMetadata
from weasyprint.fonts import FontConfiguration

from marion import __version__ as marion_version

from .. import defaults
from ..exceptions import (
    DocumentIssuerContextQueryValidationError,
    DocumentIssuerContextValidationError,
    DocumentIssuerMissingContext,
    DocumentIssuerMissingContextQuery,
)
from ..utils import static_file_fetcher


class PDFFileMetadataMixin:
    """PDF file metadata"""

    attachments = None
    authors = None
    description = None
    generator = f"Marion, version {marion_version}"
    keywords = None
    title = None

    def __init__(self):

        self._created = None
        self._metadata = None
        self._modified = None

    @property
    def metadata(self):
        """Get document PDF file metadata.

        Return default (or set) document file metadata as a
        weasyprint.document.DocumentMetadata instance.

        """
        if self._metadata is None:
            self._metadata = DocumentMetadata(
                attachments=self.get_attachments(),
                authors=self.get_authors(),
                created=self.created,
                description=self.get_description(),
                generator=self.get_generator(),
                keywords=self.get_keywords(),
                modified=self.modified,
                title=self.get_title(),
            )
        return self._metadata

    def get_attachments(self):
        """Get document PDF file attachments metadata"""
        return self.attachments

    def get_authors(self):
        """Get document PDF file 'authors' metadata"""
        return self.authors

    @property
    def created(self):
        """Get document PDF file 'created' metadata"""
        if self._created is None:
            self._created = timezone.now().isoformat()
        return self._created

    def get_description(self):
        """Get document PDF file 'description' metadata"""
        return self.description

    def get_generator(self):
        """Get document PDF file 'generator' metadata"""
        return self.generator

    def get_keywords(self):
        """Get document PDF file 'keywords' metadata"""
        return self.keywords

    @property
    def modified(self):
        """Get document PDF file 'modified' metadata"""
        if self._modified is None:
            self._modified = timezone.now().isoformat()
        return self._modified

    def get_title(self):
        """Get document PDF file 'title' metadata"""
        return self.title


# pylint: disable=not-callable
class AbstractDocument(PDFFileMetadataMixin, ABC):
    """Base document interface.

    To define a new document, one should inherit from this interface and
    implement the `fetch_context` abstract method.

    """

    # Templates
    css_template_path = None
    html_template_path = None
    template_engine = None

    # Models
    context_model: BaseModel = None
    context_query_model: BaseModel = None

    def __init__(
        self, identifier: uuid.UUID = None, context_query: Union[str, dict] = None
    ):

        # Document
        self.identifier = self.generate_identifier(identifier)
        self.document_path = self.get_document_path()

        # Data
        self.context = None
        self.context_query = (
            self.validate_context_query(context_query)
            if context_query is not None
            else None
        )

        # Templates
        self.css = None
        self.html = None

        super().__init__()

    @classmethod
    def validate_context(cls, context: Union[str, dict]) -> BaseModel:
        """Use required context pydantic model to validate input context."""

        if cls.context_model is None:
            raise DocumentIssuerMissingContext(str(_("Context model is missing")))

        try:
            if isinstance(context, str):
                context = cls.context_model.parse_raw(context)
            elif isinstance(context, dict):
                context = cls.context_model(**context)
        except ValidationError as error:
            raise DocumentIssuerContextValidationError(
                _(f"Document issuer context string is not valid: {error}")
            ) from error
        return context

    @classmethod
    def validate_context_query(cls, context_query: Union[str, dict]) -> BaseModel:
        """Use required context query pydantic model to validate input context query."""

        if cls.context_query_model is None:
            raise DocumentIssuerMissingContextQuery(
                str(_("Context query model is missing"))
            )

        try:
            if isinstance(context_query, str):
                context_query = cls.context_query_model.parse_raw(context_query)
            if isinstance(context_query, dict):
                context_query = cls.context_query_model(**context_query)
        except ValidationError as error:
            raise DocumentIssuerContextQueryValidationError(
                _(f"Document issuer context query string is not valid: {error}")
            ) from error
        return context_query

    @cached_property
    def __default_template_basename(self):
        """Get default template base name given its class name.

        Example subsequent class name transformations:
          DummyDocument
            -> _Dummy_Document
            -> Dummy_Document
            -> dummy_document
            -> dummy
        """

        return (
            re_camel_case.sub(r"_\1", self.__class__.__name__)
            .strip("_")
            .lower()
            .replace("_document", "")
        )

    def __get_template(self, template_path):
        """Get a template from its relative path.

        This method always tries to return a template using the default
        template engine if it is not set.

        """
        if self.template_engine is None:
            template_engine = self.get_template_engine()
        return template_engine.get_template(template_path)

    def generate_identifier(self, identifier=None):
        """Generate the document identifier.

        If the identifier has been set or is provided as an argument, it will
        be returned, or else a new UUID is generated.

        """

        if hasattr(self, "identifier") and self.identifier is not None:
            return self.identifier
        if identifier is not None:
            return str(identifier)
        return str(uuid.uuid4())

    def get_document_path(self):
        """Get (generated) document path.

        Return default (or set) document path as a pathlib.Path object.

        """

        if hasattr(self, "document_path") and self.document_path is not None:
            return self.document_path
        return defaults.DOCUMENTS_ROOT.joinpath(f"{self.identifier}.pdf")

    def get_document_url(self, host=None, schema="https"):
        """Get (generated) document URL.

        If the host argument is provided a fully qualified URL will be
        returned, or else, an absolute URL will be generated.

        """
        relative_path = self.get_document_path().relative_to(defaults.DOCUMENTS_ROOT)
        relative_url = f"{settings.MEDIA_URL}{relative_path}"
        if host is None:
            return relative_url
        return f"{schema}://{host}{relative_url}"

    def get_css(self):
        """Get CSS template instance"""

        if self.css is not None:
            return self.css
        return self.__get_template(self.get_css_template_path())

    def get_css_template_path(self):
        """Get CSS template path.

        Return default (or set) CSS template path as a pathlib.Path object.

        """

        if self.css_template_path is not None:
            return self.css_template_path
        return defaults.DOCUMENTS_TEMPLATE_ROOT.joinpath(
            f"{self.__default_template_basename}.css"
        )

    def get_html(self):
        """Get HTML template instance"""

        if self.html is not None:
            return self.html
        return self.__get_template(self.get_html_template_path())

    def get_html_template_path(self):
        """Get HTML template path.

        Return default (or set) HTML template path as a pathlib.Path object.

        """

        if self.html_template_path is not None:
            return self.html_template_path
        return defaults.DOCUMENTS_TEMPLATE_ROOT.joinpath(
            f"{self.__default_template_basename}.html"
        )

    def get_template_engine(self):
        """Get template engine.

        Return default (or set) template engine.

        """
        if self.template_engine is not None:
            return self.template_engine
        return Engine.get_default()

    @abstractmethod
    def fetch_context(self) -> dict:
        """Fetch document context given context query parameters.

        This method should be implemented while using this interface for a
        custom document class.

        Note that it is highly recommended to validate the context_query using
        the validate_context_query method in your implementation to ensure data
        consistency.

        Returns fetched context as a dictionnary.

        """

    def set_context(self, context: dict):
        """Validate and set context passed as a dictionary instance"""
        self.context = self.validate_context(context)

    def get_django_context(self) -> Context:
        """Get the Django Context instance from the context model instance."""
        return Context(self.context.dict())

    def create(self):
        """Create document.

        Given an HTML template, a CSS template and the required context to
        compile them, we render the document HTML that will be used by
        Weasyprint's headless web browser to generate the document as a PDF
        file.

        The path of the document is returned as a pathlib.Path instance.

        """

        if self.context is None:
            self.set_context(self.fetch_context())

        document_path = self.get_document_path()
        django_context = self.get_django_context()
        html_str = self.get_html().render(django_context)
        css_str = self.get_css().render(django_context)

        font_config = FontConfiguration()
        html = HTML(string=html_str, url_fetcher=static_file_fetcher)
        css = CSS(string=css_str, font_config=font_config)

        document = html.render(stylesheets=[css], font_config=font_config)
        document.metadata = self.metadata
        document.write_pdf(target=document_path, zoom=1)

        return document_path



"""Admin for the marion application"""

from django.contrib import admin

from .models import DocumentRequest


class DocumentRequestAdmin(admin.ModelAdmin):
    """DocumentModel admin"""


admin.site.register(DocumentRequest, DocumentRequestAdmin)



"""Tests for the marion application models"""

from django.core.exceptions import FieldError as DjangoFieldError
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import models as django_models

import pytest
from pydantic import BaseModel

from marion import exceptions, factories, issuers, models


def test_pydantic_model_field_validation():
    """Test the pydanticmodel validation for a PydanticModelField"""
    # pylint: disable=missing-class-docstring

    # Missing schema attribute and field schema extraction method
    class TestModelA(django_models.Model):
        data = models.PydanticModelField()

    instance = TestModelA(data={"fullname": "Richie"})

    with pytest.raises(
        DjangoFieldError, match="A pydantic model is missing for the 'data' field."
    ):
        instance.full_clean()

    # With a schema attribute
    class PydanticModelA(BaseModel):
        fullname: str

    class TestModelB(django_models.Model):
        data = models.PydanticModelField(pydantic_model=PydanticModelA)

    instance = TestModelB(data={"fullname": "Richie"})
    instance.full_clean()

    with pytest.raises(DjangoValidationError, match="none is not an allowed value"):
        instance = TestModelB(data={"fullname": None})
        instance.full_clean()

    # Missing schema attribute but with field schema extraction method
    class TestModelC(django_models.Model):
        data = models.PydanticModelField()

        def get_data_pydantic_model(self):
            # pylint: disable=no-self-use,missing-function-docstring
            return PydanticModelA

    instance = TestModelC(data={"fullname": "Richie"})
    instance.full_clean()

    with pytest.raises(DjangoValidationError, match="none is not an allowed value"):
        instance = TestModelC(data={"fullname": None})
        instance.full_clean()

    # If both schema attribute and schema extraction method are provided, the
    # schema attribute should prevail
    class PydanticModelB(BaseModel):
        amount: int

        class Config:
            extra = "forbid"

    class TestModelD(django_models.Model):
        data = models.PydanticModelField(pydantic_model=PydanticModelB)

        def get_data_pydantic_model(self):
            # pylint: disable=no-self-use,missing-function-docstring
            return PydanticModelB

    instance = TestModelD(data={"amount": 2})
    instance.full_clean()

    with pytest.raises(
        DjangoValidationError,
        match="extra fields not permitted",
    ):
        instance = TestModelD(data={"fullname": "Richie"})
        instance.full_clean()


@pytest.mark.django_db
def test_document_request_default_ordering():
    """Test the `DocumentRequest` default ordering"""

    first_document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )
    second_document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Marion Cunningham"},
    )
    third_document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Howard Cunningham"},
    )
    fourth_document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Joanie Cunningham"},
    )

    document_requests = models.DocumentRequest.objects.all()
    assert document_requests[0] == fourth_document_request
    assert document_requests[1] == third_document_request
    assert document_requests[2] == second_document_request
    assert document_requests[3] == first_document_request


@pytest.mark.django_db
def test_document_request_save(monkeypatch):
    """Test the `DocumentRequest.save()` method"""

    document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )

    assert document_request.context_query.get("fullname") == "Richie Cunningham"

    # Test fetched context
    assert document_request.context.get("fullname") == "Richie Cunningham"
    assert "identifier" in document_request.context
    assert document_request.context.get("identifier") is not None

    # Test created document
    assert document_request.context.get("identifier") == str(
        document_request.document_id
    )
    assert document_request.id != document_request.document_id
    assert document_request.get_issuer().get_document_path().exists()

    # We should perform validation before saving
    document_request = factories.DocumentRequestFactory.build(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": ""},
    )
    with pytest.raises(
        DjangoValidationError,
        match="ensure this value has at least 2 characters",
    ):
        document_request.save()

    document_request = factories.DocumentRequestFactory.build(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )
    with monkeypatch.context() as mk_patch:
        mk_patch.setattr(
            issuers.DummyDocument,
            "fetch_context",
            lambda *args, **kwargs: {"fullname": "Richie Cunningham"},
        )
        with pytest.raises(
            exceptions.DocumentIssuerContextValidationError,
            match="identifier\n  field required",
        ):
            document_request.save()


def test_document_request_get_issuer_class(monkeypatch):
    """Test the `DocumentRequest.get_issuer_class()` method"""

    # pylint: disable=no-member
    monkeypatch.setattr(
        models.DocumentRequest.issuer.field,
        "choices",
        [
            ("marion.issuers.Arnold", "Arnold"),
            ("marion.issuers.DummyDocument", "Dummy"),
            ("marion.issuers.Marsha", "Marsha"),
            ("marion.issuers.Richie", "Richie"),
            ("howard.documents.issuers.Richie", "Richou"),
            ("issuers.Richie", "Richinou"),
            ("Richie", "Riccardo"),
        ],
    )

    with pytest.raises(
        exceptions.InvalidDocumentIssuer, match="Fonzy is not an allowed issuer"
    ):
        models.DocumentRequest.get_issuer_class("Fonzy")

    with pytest.raises(
        exceptions.InvalidDocumentIssuer,
        match="Issuer class name should be unique, found 4 for Richie",
    ):
        models.DocumentRequest.get_issuer_class("Richie")

    assert isinstance(
        models.DocumentRequest.get_issuer_class("DummyDocument")(),
        issuers.DummyDocument,
    )


@pytest.mark.django_db
def test_document_request_get_issuer():
    """Test the `DocumentRequest.get_issuer()` method"""

    document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )

    assert isinstance(document_request.get_issuer(), issuers.DummyDocument)


@pytest.mark.django_db
def test_document_request_get_document_url():
    """Test the `DocumentRequest.get_document_url()` method"""

    document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )

    assert (
        document_request.get_document_url()
        == f"/media/{document_request.document_id}.pdf"
    )
    assert (
        document_request.get_document_url(host="www.happy-days.com")
        == f"https://www.happy-days.com/media/{document_request.document_id}.pdf"
    )
    assert (
        document_request.get_document_url(host="www.happy-days.com", schema="http")
        == f"http://www.happy-days.com/media/{document_request.document_id}.pdf"
    )



"""Tests for the marion application serializers"""

from django.urls import reverse

import pytest
from rest_framework.test import APIRequestFactory

from marion import factories, serializers


@pytest.mark.django_db
def test_document_request_serializer_document_url_field():
    """Test the document request serializer document_url field"""

    document_request = factories.DocumentRequestFactory(
        issuer="marion.issuers.DummyDocument",
        context_query={"fullname": "Richie Cunningham"},
    )

    factory = APIRequestFactory()
    request = factory.get(reverse("documentrequest-list"))

    serialized_document_request = serializers.DocumentRequestSerializer(
        document_request, context={"request": request}
    )

    assert (
        serialized_document_request.data.get("document_url")
        == f"http://testserver/media/{document_request.document_id}.pdf"
    )



# Generated by Django 3.1.7 on 2021-03-19 18:08

import uuid

from django.db import migrations, models

import marion.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DocumentRequest",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        help_text="Primary key for the document request as UUID",
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_on",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="Date and time at which a document request was created",
                        verbose_name="Created on",
                    ),
                ),
                (
                    "updated_on",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="Date and time at which a document request was last updated",
                        verbose_name="Updated on",
                    ),
                ),
                (
                    "document_id",
                    models.UUIDField(
                        help_text="Generated document identifier",
                        null=True,
                        unique=True,
                        verbose_name="Document ID",
                    ),
                ),
                (
                    "issuer",
                    models.CharField(
                        choices=[
                            ("howard.issuers.RealisationCertificate", "Realisation")
                        ],
                        help_text="The issuer of the document among allowed ones",
                        max_length=200,
                        verbose_name="Issuer",
                    ),
                ),
                (
                    "context",
                    marion.models.PydanticModelField(
                        blank=True,
                        editable=False,
                        help_text="Context used to render the document's template",
                        null=True,
                        verbose_name="Collected context",
                    ),
                ),
                (
                    "context_query",
                    marion.models.PydanticModelField(
                        help_text="Context will be fetched from those parameters",
                        verbose_name="Context query parameters",
                    ),
                ),
            ],
            options={
                "ordering": ["-created_on"],
            },
        ),
    ]



"""Urls for the marion application"""

from django.urls import include, path

from rest_framework import routers

from .. import views

router = routers.DefaultRouter()
router.register(r"requests", views.DocumentRequestViewSet)

urlpatterns = [
    path("", include(router.urls)),
]



"""Tests for the marion application views"""

import json
import tempfile
from pathlib import Path

from django.urls import reverse

import pytest
from pytest_django import asserts as django_assertions
from rest_framework import exceptions as drf_exceptions
from rest_framework import status
from rest_framework.test import APIClient

from marion import defaults, models
from marion.issuers import DummyDocument

client = APIClient()


def count_documents(root):
    """Return the number of generated PDF files in the root directory"""

    return len(list(root.glob("*.pdf")))


@pytest.mark.django_db
def test_document_request_viewset_post(monkeypatch):
    """Test the DocumentRequestViewSet create view"""

    monkeypatch.setattr(defaults, "DOCUMENTS_ROOT", Path(tempfile.mkdtemp()))

    url = reverse("documentrequest-list")

    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Request payload required parameters
    data = {}
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert isinstance(response.data.get("context_query")[0], drf_exceptions.ErrorDetail)
    assert response.data.get("context_query")[0].code == "required"
    assert isinstance(response.data.get("issuer")[0], drf_exceptions.ErrorDetail)
    assert response.data.get("issuer")[0].code == "required"
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Invalid issuer
    data = {
        "issuer": "marion.issuers.DumberDocument",
        "context_query": json.dumps({"fullname": "Richie Cunningham"}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert response.data.get("issuer")[0].code == "invalid_choice"
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Perform standard request
    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": "Richie Cunningham"}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_201_CREATED
    assert models.DocumentRequest.objects.count() == 1
    assert (
        models.DocumentRequest.objects.get().context.get("fullname")
        == "Richie Cunningham"
    )
    assert count_documents(defaults.DOCUMENTS_ROOT) == 1


@pytest.mark.django_db
def test_document_request_viewset_post_context_query_pydantic_model_validation(
    monkeypatch,
):
    """Test the DocumentRequestViewSet create view context_query pydantic model
    validation.

    """

    monkeypatch.setattr(defaults, "DOCUMENTS_ROOT", Path(tempfile.mkdtemp()))

    url = reverse("documentrequest-list")

    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Refuse extra fields in context query
    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": "Richie Cunningham", "friends": 2}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "extra fields not permitted" in str(response.data.get("error"))
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Input types checking
    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": None}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "none is not an allowed value" in str(response.data.get("error"))
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Input contraints checking (short fullname)
    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": "D"}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "ensure this value has at least 2 characters" in str(
        response.data.get("error")
    )
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Input contraints checking (too long fullname)
    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": "F" * 256}),
    }
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "ensure this value has at most 255 characters" in str(
        response.data.get("error")
    )
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0


@pytest.mark.django_db
def test_document_request_viewset_post_context_pydantic_model_validation(
    monkeypatch,
):
    """Test the DocumentRequestViewSet create view context pydantic model
    validation.

    """
    # pylint: disable=unused-argument,function-redefined

    monkeypatch.setattr(defaults, "DOCUMENTS_ROOT", Path(tempfile.mkdtemp()))

    url = reverse("documentrequest-list")

    data = {
        "issuer": "marion.issuers.DummyDocument",
        "context_query": json.dumps({"fullname": "Richie Cunningham"}),
    }

    # Refuse extra fields in context
    def mock_fetch_context(*args, **kwargs):
        """A mock that returns invalid context"""
        return {
            "fullname": "Richie Cunningham",
            "identifier": "0a1c3ccf-c67d-4071-ab1f-3b27628db9b1",
            "friends": 2,
        }

    monkeypatch.setattr(DummyDocument, "fetch_context", mock_fetch_context)
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "extra fields not permitted" in response.data.get("error")
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Types checking
    def mock_fetch_context(*args, **kwargs):
        """A mock that returns invalid context"""
        return {"fullname": None, "identifier": "0a1c3ccf-c67d-4071-ab1f-3b27628db9b1"}

    monkeypatch.setattr(DummyDocument, "fetch_context", mock_fetch_context)
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "none is not an allowed value" in response.data.get("error")
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Missing identifier
    def mock_fetch_context(*args, **kwargs):
        """A mock that returns invalid context"""
        return {"fullname": "Richie Cunningham"}

    monkeypatch.setattr(DummyDocument, "fetch_context", mock_fetch_context)
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "identifier\n  field required" in response.data.get("error")
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Constraints checking (short fullname)
    def mock_fetch_context(*args, **kwargs):
        """A mock that returns invalid context"""
        return {"fullname": "D", "identifier": "0a1c3ccf-c67d-4071-ab1f-3b27628db9b1"}

    monkeypatch.setattr(DummyDocument, "fetch_context", mock_fetch_context)
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "ensure this value has at least 2 characters" in response.data.get("error")
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0

    # Constraints checking (too long fullname)
    def mock_fetch_context(*args, **kwargs):
        """A mock that returns invalid context"""
        return {
            "fullname": "F" * 256,
            "identifier": "0a1c3ccf-c67d-4071-ab1f-3b27628db9b1",
        }

    monkeypatch.setattr(DummyDocument, "fetch_context", mock_fetch_context)
    response = client.post(url, data, format="json")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "ensure this value has at most 255 characters" in response.data.get("error")
    assert models.DocumentRequest.objects.count() == 0
    assert count_documents(defaults.DOCUMENTS_ROOT) == 0


def test_document_template_debug_view_is_only_active_in_debug_mode(settings):
    """Test if the document_template_debug view is active when not in debug mode"""

    settings.DEBUG = False
    url = reverse("documents-template-debug")

    response = client.get(url)
    assert response.status_code == 403


def test_document_template_debug_view(settings):
    """Test the document_template_debug view"""

    settings.DEBUG = True
    settings.MARION_DOCUMENT_ISSUER_CHOICES_CLASS = (
        "marion.default.DocumentIssuerChoices"
    )
    url = reverse("documents-template-debug")

    response = client.get(url)
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert b"You should provide an issuer." in response.content

    response = client.get(url, {"issuer": "foo.bar.baz"})
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert b"Unknown issuer foo.bar.baz" in response.content

    response = client.get(url, {"issuer": "marion.issuers.DummyDocument"})
    assert response.status_code == 200
    # pylint: disable=no-member
    django_assertions.assertContains(response, "<h1>Dummy document</h1>")



"""Tests for the marion.issuers.base document"""

import uuid
from datetime import datetime
from pathlib import Path
from unittest.mock import patch

from django.template import Context, Template, engines
from django.template.engine import Engine

import pytest
from pdfminer.high_level import extract_text as pdf_extract_text
from pydantic import BaseModel
from weasyprint.document import DocumentMetadata

from marion.defaults import DOCUMENTS_ROOT
from marion.exceptions import (
    DocumentIssuerContextQueryValidationError,
    DocumentIssuerContextValidationError,
    DocumentIssuerMissingContext,
    DocumentIssuerMissingContextQuery,
)
from marion.issuers.base import AbstractDocument


def test_abstract_document_interface_with_missing_abstract_methods():
    """Test interface mechanism with missing abstract methods"""

    class BadTestDocument(AbstractDocument):
        """Bad implementation with missing abstract methods"""

    # pylint: disable=abstract-class-instantiated
    with pytest.raises(
        TypeError,
        match=(
            "Can't instantiate abstract class BadTestDocument with "
            "abstract methods fetch_context"
        ),
    ):
        BadTestDocument()


def test_abstract_document_interface_with_implemented_abstract_method():
    """Test interface mechanism with properly implemented abstract methods"""

    class GoodTestDocument(AbstractDocument):
        """Correct implementation with required abstract methods"""

        def fetch_context(self, **context_query):
            """Fake context fetching"""

    GoodTestDocument()


def test_abstract_document_init(monkeypatch):
    """Test AbstractDocument init method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()

    freezed_now = datetime(2021, 1, 1, 0, 0, 0)
    monkeypatch.setattr("django.utils.timezone.now", lambda: freezed_now)

    # Instance attributes
    assert test_document.identifier is not None
    assert isinstance(test_document.identifier, str)
    assert test_document.document_path is not None
    assert isinstance(test_document.document_path, Path)
    assert test_document.context is None
    assert test_document.css is None
    assert test_document.html is None

    # PDFFileMetadataMixin attributes
    assert test_document.created == freezed_now.isoformat()
    assert test_document.metadata is not None
    assert isinstance(test_document.metadata, DocumentMetadata)
    assert test_document.modified == freezed_now.isoformat()


def test_abstract_document_generate_identifier():
    """Test AbstractDocument generate_identifier method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.generate_identifier() == test_document.identifier

    identifier = uuid.uuid4()
    test_document = TestDocument(identifier=str(identifier))
    assert test_document.generate_identifier() == str(identifier)

    other_identifier = uuid.uuid4()
    assert test_document.generate_identifier(str(other_identifier)) == str(identifier)

    test_document.identifier = None
    assert test_document.generate_identifier(str(other_identifier)) == str(
        other_identifier
    )


def test_abstract_document_get_document_path():
    """Test AbstractDocument get_document_path method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_document_path() == Path(
        f"{DOCUMENTS_ROOT}/{test_document.identifier}.pdf"
    )

    test_document.document_path = Path(f"{DOCUMENTS_ROOT}/richie.pdf")
    assert test_document.get_document_path() == Path(f"{DOCUMENTS_ROOT}/richie.pdf")


def test_abstract_document_get_document_url():
    """Test AbstractDocument get_document_url method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_document_url() == f"/media/{test_document.identifier}.pdf"
    assert (
        test_document.get_document_url(host="example.org")
        == f"https://example.org/media/{test_document.identifier}.pdf"
    )
    assert (
        test_document.get_document_url(host="example.org", schema="http")
        == f"http://example.org/media/{test_document.identifier}.pdf"
    )


def test_abstract_document_get_css(monkeypatch):
    """Test AbstractDocument get_css method"""

    monkeypatch.setattr(
        Engine, "get_template", lambda _, template_name: "body { color: red }"
    )

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_css() is not None

    # Ensure we always return css attribute when set
    test_document.css = "body { color: red }"
    assert test_document.get_css() == "body { color: red }"
    monkeypatch.setattr(
        Engine, "get_template", lambda _, template_name: "body { color: blue }"
    )
    assert test_document.get_css() == "body { color: red }"


def test_abstract_document_get_css_template_path():
    """Test AbstractDocument get_css_template_path method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_css_template_path() == Path("marion/test.css")

    test_document.css_template_path = Path("foo/bar.css")
    assert test_document.get_css_template_path() == Path("foo/bar.css")

    class TestDocumentWithSpecificHTMLTemplatePath(AbstractDocument):
        css_template_path = Path("foo/specific.css")

        def fetch_context(self, **context_query):
            pass

    test_document = TestDocumentWithSpecificHTMLTemplatePath()
    assert test_document.get_css_template_path() == Path("foo/specific.css")


def test_abstract_document_get_html(monkeypatch):
    """Test AbstractDocument get_html method"""

    monkeypatch.setattr(
        Engine, "get_template", lambda _, template_name: "body { color: red }"
    )

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_html() is not None

    # Ensure we always return html attribute when set
    test_document.html = "body { color: red }"
    assert test_document.get_html() == "body { color: red }"
    monkeypatch.setattr(
        Engine, "get_template", lambda _, template_name: "body { color: blue }"
    )
    assert test_document.get_html() == "body { color: red }"


def test_abstract_document_get_html_template_path():
    """Test AbstractDocument get_html_template_path method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    assert test_document.get_html_template_path() == Path("marion/test.html")

    test_document.html_template_path = Path("foo/bar.html")
    assert test_document.get_html_template_path() == Path("foo/bar.html")

    # pylint: disable=missing-class-docstring
    class TestDocumentWithSpecificHTMLTemplatePath(AbstractDocument):
        html_template_path = Path("foo/specific.html")

        def fetch_context(self, **context_query):
            pass

    test_document = TestDocumentWithSpecificHTMLTemplatePath()
    assert test_document.get_html_template_path() == Path("foo/specific.html")


def test_abstract_document_get_template_engine():
    """Test AbstractDocument get_template_engine method"""

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()
    with patch.object(Engine, "get_default") as mocked_get_default:
        test_document.get_template_engine()
        mocked_get_default.assert_called()

    test_document.template_engine = "fake"
    with patch.object(Engine, "get_default") as mocked_get_default:
        test_document.get_template_engine()
        mocked_get_default.assert_not_called()


def test_abstract_document_validate_context():
    """Test AbstractDocument validate_context method"""

    # pylint: disable=missing-class-docstring
    class TestDocumentWithMissingContextModel(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    with pytest.raises(DocumentIssuerMissingContext, match="Context model is missing"):
        TestDocumentWithMissingContextModel.validate_context({"foo": 1})

    class ContextModel(BaseModel):
        fullname: str
        friends: int

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):

        context_model = ContextModel

        def fetch_context(self, **context_query):
            pass

    with pytest.raises(
        DocumentIssuerContextValidationError, match="value is not a valid integer"
    ):
        TestDocument.validate_context(
            {"fullname": "Richie Cunningham", "friends": "None"}
        )

    assert TestDocument.validate_context(
        {"fullname": "Richie Cunningham", "friends": 2}
    ) == ContextModel(
        fullname="Richie Cunningham",
        friends=2,
    )


def test_abstract_document_validate_context_query():
    """Test AbstractDocument validate_context_query method"""

    # pylint: disable=missing-class-docstring
    class TestDocumentWithMissingContextModel(AbstractDocument):
        def fetch_context(self, **context_query):
            pass

    with pytest.raises(
        DocumentIssuerMissingContextQuery,
        match="Context query model is missing",
    ):
        TestDocumentWithMissingContextModel.validate_context_query({"foo": 1})

    # pylint: disable=missing-class-docstring
    class ContextQueryModel(BaseModel):
        fullname: str
        friends: int

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):

        context_query_model = ContextQueryModel

        def fetch_context(self, **context_query):
            pass

    with pytest.raises(
        DocumentIssuerContextQueryValidationError,
        match="value is not a valid integer",
    ):
        TestDocument.validate_context_query(
            {"fullname": "Richie Cunningham", "friends": "None"}
        )

    assert TestDocument.validate_context_query(
        {"fullname": "Richie Cunningham", "friends": 2}
    ) == ContextQueryModel(
        fullname="Richie Cunningham",
        friends=2,
    )


def test_abstract_document_set_context():
    """Test AbstractDocument set_context method"""

    # pylint: disable=missing-class-docstring
    class ContextModel(BaseModel):
        fullname: str
        friends: int

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):

        context_model = ContextModel

        def fetch_context(self, **context_query):
            pass

    test_document = TestDocument()

    # Load valid context
    context = Context({"fullname": "Richie Cunningham", "friends": 2})
    test_document.set_context(context)
    assert test_document.context == context

    # Load invalid context
    context = {"fullname": "Richie Cunningham", "friends": "None"}
    with pytest.raises(
        DocumentIssuerContextValidationError,
        match="value is not a valid integer",
    ):
        test_document.set_context(context)


# pylint: disable=missing-class-docstring
def test_abstract_document_get_django_context():
    """Test AbstractDocument get_django_context method"""

    class ContextModel(BaseModel):
        life: int

    class ContextQueryModel(BaseModel):
        intergalactic: str

    class TestDocument(AbstractDocument):
        context_model = ContextModel
        context_query_model = ContextQueryModel

        def fetch_context(self):
            return {"life": 42}

    document = TestDocument(context_query={"intergalactic": "voyager"})
    document.set_context(document.fetch_context())
    assert document.get_django_context() == Context({"life": 42})


def test_abstract_document_create():
    """Test AbstractDocument create method"""

    # pylint: disable=missing-class-docstring
    class ContextModel(BaseModel):
        user_id: str
        fullname: str

    # pylint: disable=missing-class-docstring
    class ContextQueryModel(BaseModel):
        user_id: str

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        context_model = ContextModel
        context_query_model = ContextQueryModel

        def get_html(self):
            return Template(
                "<body>My name is {{ fullname }} (user id: {{ user_id }})</body>"
            )

        def get_css(self):
            return Template("body {color: red}")

        def fetch_context(self):
            return {"fullname": "Richie Cunningham", **self.context_query.dict()}

    test_document = TestDocument(context_query={"user_id": "rcunningham"})
    test_document_path = test_document.create()

    assert Path(test_document_path).exists()
    with test_document_path.open("rb") as test_document_file:
        assert (
            pdf_extract_text(test_document_file).strip()
            == "My name is Richie Cunningham (user id: rcunningham)"
        )


def test_abstract_document_jinja_template_engine(settings):
    """Test document rendering with jinja templates"""

    settings.TEMPLATES = [
        {"BACKEND": "django.template.backends.django.DjangoTemplates", "DIRS": []},
        {"BACKEND": "django.template.backends.jinja2.Jinja2", "DIRS": []},
    ]

    # pylint: disable=missing-class-docstring
    class TestDocument(AbstractDocument):
        template_engine = engines["jinja2"]

        def get_html(self):
            return self.template_engine.from_string(
                "{% set status='ninja' %}"
                "<body>My name is {{ fullname }} (status: {{ status }})</body>"
            )

        def fetch_context(self):
            pass

    test_document = TestDocument()
    assert (
        test_document.get_html().render({"fullname": "Richie Cunningham"})
        == "<body>My name is Richie Cunningham (status: ninja)</body>"
    )



"""Models for the marion application"""

import json
import uuid

from django.core.exceptions import FieldError
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import models
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from pydantic.error_wrappers import ValidationError as PydanticValidationError

from .defaults import DOCUMENT_ISSUER_CHOICES_CLASS
from .exceptions import DocumentIssuerContextQueryValidationError, InvalidDocumentIssuer

DocumentIssuerChoices = import_string(DOCUMENT_ISSUER_CHOICES_CLASS)


class PydanticModelField(models.JSONField):
    """Pydantic Model Field.

    This field is a pydantic model field but with model validation when the model
    is provided.

    """

    def __init__(self, *args, **kwargs):
        self.pydantic_model = kwargs.pop("pydantic_model", None)
        super().__init__(*args, **kwargs)

    def _validate_pydantic_model(self, value, model_instance):
        """Perform pydantic model validation"""

        pydantic_model = self._get_pydantic_model(model_instance)

        # Disable validation when migrations are faked
        if self.model.__module__ == "__fake__":
            return

        # Validate either raw (JSON string) data or a serialized dict (before
        # saving the Django model).
        try:
            if isinstance(value, str):
                pydantic_model.parse_raw(value)
            elif isinstance(value, dict):
                pydantic_model(**value)
        except PydanticValidationError as error:
            raise DjangoValidationError(error, code="invalid") from error

    def _get_pydantic_model(self, model_instance):
        """Get field pydantic model from the expected model instance method"""

        if self.pydantic_model is None:
            try:
                self.pydantic_model = getattr(
                    model_instance, f"get_{self.name}_pydantic_model"
                )()
            except AttributeError as error:
                raise FieldError(
                    _(
                        f"A pydantic model is missing for the '{self.name}' field. "
                        "It should be provided thanks to the 'pydantic_model' "
                        "field argument or by adding a get_<FIELD_NAME>_pydantic_model "
                        "method to your model."
                    )
                ) from error
        return self.pydantic_model

    def validate(self, value, model_instance):
        """Add pydantic model validation to field validation"""

        # Validate JSON value
        super().validate(value, model_instance)

        self._validate_pydantic_model(value, model_instance)

    def pre_save(self, model_instance, add):
        """Ensure pydantic model validation occurs before saving"""

        value = super().pre_save(model_instance, add)
        if value and not self.null:
            self._validate_pydantic_model(value, model_instance)
        return value


class DocumentRequest(models.Model):
    """Document Requests are stored in models and linked to documents.

    Each document request has a particular issuer, context, and
    context_query. An issuer is responsible for the generation of a particular
    type of document in PDF format; it has its own logic that implements how
    to fetch required context to compile a document template given a
    context_query.

    Note that a document request has its own unique identifier that is
    distinct from the document unique identifier as we consider them as
    dictinct objects; but we make sure that one cannot generate two
    documents with the same identifier from two document requests.
    Another advantage of having a distinct identifier for a document request
    and a request is that one cannot guess a document request identifier
    from a document identifier, and thus one cannot access to the
    document request object via the document requests API.

    Given an issuer and a context, we should be able to re-generate a
    particular document if the PDF file has been lost or deleted. In last
    resort, we also should be able to re-generate the context from the
    context_query, but the latest is mostly stored for tracking purpose.

    """

    id = models.UUIDField(
        verbose_name=_("ID"),
        help_text=_("Primary key for the document request as UUID"),
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )

    created_on = models.DateTimeField(
        verbose_name=_("Created on"),
        help_text=_("Date and time at which a document request was created"),
        auto_now_add=True,
        editable=False,
    )

    updated_on = models.DateTimeField(
        verbose_name=_("Updated on"),
        help_text=_("Date and time at which a document request was last updated"),
        auto_now=True,
        editable=False,
    )

    document_id = models.UUIDField(
        verbose_name=_("Document ID"),
        help_text=_("Generated document identifier"),
        null=True,
        unique=True,
    )

    issuer = models.CharField(
        verbose_name=_("Issuer"),
        help_text=_("The issuer of the document among allowed ones"),
        max_length=200,
        choices=DocumentIssuerChoices.choices,
    )

    context = PydanticModelField(
        pydantic_model=None,
        verbose_name=_("Collected context"),
        help_text=_("Context used to render the document's template"),
        editable=False,
        null=True,
        blank=True,
    )

    context_query = PydanticModelField(
        pydantic_model=None,
        verbose_name=_("Context query parameters"),
        help_text=_("Context will be fetched from those parameters"),
    )

    class Meta:
        """Options for the DocumentRequest model"""

        ordering = ["-created_on"]

    def get_context_pydantic_model(self):
        """Get context pydantic model from the issuer class"""
        return self.get_issuer_class(self.issuer).context_model

    def get_context_query_pydantic_model(self):
        """Get context_query pydantic model from the issuer class"""
        return self.get_issuer_class(self.issuer).context_query_model

    # pylint: disable=signature-differs
    def save(self, *args, **kwargs):
        """Generate the document along with the document request"""

        document = self.get_issuer()
        document.create()

        self.document_id = document.identifier

        # Prevent JSON encoding issues
        #
        # Pydantic knows how to JSON-serialize all fields, the standard JSON
        # encoder does not. So we convert pydantic model data to a
        # dumb-dict with simple types using this trick.
        self.context = json.loads(document.context.json())
        self.context_query = json.loads(document.context_query.json())

        super().save(*args, **kwargs)

    @classmethod
    def get_issuer_class(cls, issuer_class_name):
        """Get issuer class given its class name (or its path)"""

        # pylint: disable=no-member
        issuer_class_paths = list(
            filter(
                lambda issuer_module_path: issuer_module_path.endswith(
                    issuer_class_name
                ),
                # pylint: disable=no-member
                (issuer for issuer, _ in cls.issuer.field.choices),
            )
        )
        matches = len(issuer_class_paths)
        if matches == 0:
            raise InvalidDocumentIssuer(
                _(f"{issuer_class_name} is not an allowed issuer")
            )
        if matches > 1:
            raise InvalidDocumentIssuer(
                _(
                    f"Issuer class name should be unique, found {matches} for "
                    f"{issuer_class_name}"
                )
            )
        return import_string(issuer_class_paths[0])

    def get_issuer(self):
        """Get instanciated issuer class"""

        try:
            return self.get_issuer_class(self.issuer)(
                identifier=self.document_id, context_query=self.context_query
            )
        except DocumentIssuerContextQueryValidationError as error:
            raise DjangoValidationError(error, code="invalid") from error

    def get_document_url(self, host=None, schema="https"):
        """Shortcut to get the document URL.

        This method is mostly a wrapper for the issuer's `get_document_url`
        method. If no host or schema is provided, an absolute URL is returned
        (and not a fully qualified URL).

        """
        return self.get_issuer().get_document_url(host=host, schema=schema)



"""Serializers for the marion application"""

from rest_framework import serializers

from .models import DocumentRequest


class DocumentRequestSerializer(serializers.HyperlinkedModelSerializer):
    """DocumentRequest model serializer"""

    class Meta:
        model = DocumentRequest
        fields = "__all__"

    document_url = serializers.SerializerMethodField()

    def get_document_url(self, instance):
        """Add the document URL to the object"""

        return self._context.get("request").build_absolute_uri(
            instance.get_document_url()
        )



"""Exceptions for the marion application"""


class InvalidDocumentIssuer(Exception):
    """Invalid document issuer exception.

    This exception is raised when one tries to instanciate a document issuer
    that does not exists or is not activated in the
    DOCUMENT_ISSUER_CHOICES_CLASS setting (see defaults).

    """


class DocumentIssuerContextQueryValidationError(Exception):
    """Document issuer context_query validation error.

    This exception is raised when the input context_query for a document
    issuer does not validate the document issuer context_query schema.

    """


class DocumentIssuerContextValidationError(Exception):
    """Document issuer context validation error.

    This exception is raised when the context fetched by a document issuer
    does not validate the document issuer context schema.

    """


class DocumentIssuerMissingContext(Exception):
    """Document issuer missing context error.

    This exception is raised when the document issuer context is missing to
    perform an action, e.g. to create a document.

    """


class DocumentIssuerMissingContextQuery(Exception):
    """Document issuer missing context query error.

    This exception is raised when the document issuer context query is missing
    to perform an action, e.g. to create a document.

    """
